import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:collection/collection.dart';
import 'package:path/path.dart' as path;

import '../_log.dart';
import '../_utils.dart';
import '../file_collection.dart';
import '../helpers.dart';
import '../task.dart';
import '../task_invocation.dart';

/// Current version of the Dartle Cache.
const cacheFormatVersion = '0.2';

/// Kind of [FileChange].
enum ChangeKind {
  added,
  deleted,
  modified,
}

/// FileChange represents a file system entity change.
class FileChange {
  final FileSystemEntity entity;
  final ChangeKind kind;

  FileChange(this.entity, this.kind);
}

/// The cache used by dartle to figure out when files change between checks,
/// typically between two builds.
///
/// It is a cache-based cache - it does not cache the contents of files or
/// directories, it only associates a cache to them so that it can tell whether
/// their contents have changed between two checks in a very efficient manner.
class DartleCache {
  static DartleCache? _defaultCache;

  /// Get the default DartleCache instance.
  /// May initialize the default cache directory at .dartle_tool.
  static DartleCache get instance {
    return _defaultCache ??= DartleCache._defaultInstance();
  }

  final String rootDir;
  final String _hashesDir;
  final String _tasksDir;
  final String _executablesDir;

  /// Create an instance of [DartleCache] at the given root directory.
  DartleCache(this.rootDir)
      : _hashesDir = path.join(rootDir, 'hashes'),
        _tasksDir = path.join(rootDir, 'tasks'),
        _executablesDir = path.join(rootDir, 'executables') {
    init();
  }

  DartleCache._defaultInstance() : this(dartleDir);

  /// Initialize the cache directories.
  ///
  /// This method does not normally need to be called explicitly as the
  /// constructor will call it.
  void init() {
    final root = Directory(rootDir);
    final versionFile = File(path.join(rootDir, 'version'));
    bool requiresUpdate;
    if (root.existsSync()) {
      requiresUpdate = !_isCurrentVersion(versionFile);
      if (requiresUpdate) {
        logger.info(
            'Dartle cache version change detected. Performing full cleanup.');
        Directory(_hashesDir).deleteSync(recursive: true);
        Directory(_tasksDir).deleteSync(recursive: true);
        Directory(_executablesDir).deleteSync(recursive: true);
      }
    } else {
      root.createSync(recursive: true);
      requiresUpdate = true;
    }
    if (requiresUpdate) {
      versionFile.writeAsStringSync(cacheFormatVersion);
    }
    Directory(_hashesDir).createSync();
    Directory(_tasksDir).createSync();
    Directory(_executablesDir).createSync();
    logger.fine(() =>
        'Cache initialized at ${path.join(Directory.current.path, rootDir)}');
  }

  bool _isCurrentVersion(File versionFile) {
    if (versionFile.existsSync()) {
      final version = versionFile.readAsStringSync();
      return version == cacheFormatVersion;
    } else {
      return false;
    }
  }

  /// Clean the Dartle cache.
  ///
  /// If the key is empty, then the cache is cleaned completely, otherwise only
  /// entries associated with the key are removed.
  Future<void> clean({String key = ''}) async {
    if (key.isEmpty) {
      logger.fine('Cleaning Dartle cache');
      await ignoreExceptions(
          () => Directory(_tasksDir).delete(recursive: true));
      await ignoreExceptions(
          () => Directory(_hashesDir).delete(recursive: true));
      logger.finest('Dartle cache has been cleaned');
    } else {
      logger.fine(() => 'Cleaning Dartle cache (key=$key)');
      final dir = Directory(path.join(_hashesDir, _encodeKey(key)));
      await ignoreExceptions(() => dir.delete(recursive: true));
      logger.finest(() => 'Dartle cache has been cleaned (key=$key)');
    }
  }

  /// Remove from this cache all files and directories in the given collection.
  Future<void> remove(FileCollection collection, {String key = ''}) async {
    await for (final entry in collection.resolve()) {
      await _removeEntity(entry.entity, key: key);
    }
  }

  /// Cache all files and directories in the given collection.
  Future<void> call(FileCollection collection, {String key = ''}) async {
    logger.finest(() => 'Adding $collection with key="$key" to cache');
    Set<String> visitedEntities = {};
    await for (final entry in collection.resolve()) {
      if (visitedEntities.add(entry.path)) {
        await entry.use((file) => _cacheFile(file, key: key),
            (dir, children) => _cacheDir(dir, children, key: key));
      }
    }
    // visit entities that do not exist but may have existed before
    for (final file in collection.files.where(visitedEntities.add)) {
      _removeEntity(File(file), key: key);
    }
    for (final dir in collection.directories
        .map((e) => e.path)
        .where(visitedEntities.add)) {
      _removeEntity(Directory(dir), key: key);
    }
  }

  /// Check if the given file system entity is present in the cache.
  bool contains(FileSystemEntity entity, {String key = ''}) =>
      _getCacheLocation(entity, key: key).existsSync();

  /// Cache the given task invocation.
  Future<void> cacheTaskInvocation(TaskInvocation invocation) async {
    final file =
        File(path.join(_tasksDir, invocation.name.escapePathSeparator()));
    logger.fine(() =>
        'Caching invocation of task "${invocation.name}" at ${file.path}');
    await file.writeAsString(invocation.args.toString());
  }

  /// Get the [DateTime] when this task was last invoked successfully.
  ///
  /// This time is only known if the [TaskInvocation] was previously cached via
  /// [cacheTaskInvocation].
  Future<DateTime?> getLatestInvocationTime(TaskInvocation invocation) async {
    final file =
        File(path.join(_tasksDir, invocation.name.escapePathSeparator()));
    if (await file.exists()) {
      return await file.lastModified();
    }
    return null;
  }

  /// Check if the given task had been invoked with the same arguments before.
  ///
  /// Only successful task invocations are normally cached, hence this method
  /// will normally return `true` when the previous invocation of [Task] failed.
  Future<bool> hasTaskInvocationChanged(TaskInvocation invocation) async {
    final taskFile =
        File(path.join(_tasksDir, invocation.name.escapePathSeparator()));
    if (await taskFile.exists()) {
      final taskArgs = await taskFile.readAsString();
      final isChanged = invocation.args.toString() != taskArgs;
      if (isChanged) {
        logger.fine(() => 'Task "${invocation.name}" invocation changed '
            'because args were $taskArgs, but is now ${invocation.args}.');
      } else {
        logger.finest(() => 'Task "${invocation.name}" invocation has not '
            'changed, args are $taskArgs');
      }
      return isChanged;
    } else {
      logger.fine(() => 'Task "${invocation.name}" has not been executed yet');
      return true;
    }
  }

  /// Remove any previous invocations of a task with the given name
  /// from the cache.
  Future<void> removeTaskInvocation(String taskName) async {
    final file = File(path.join(_tasksDir, taskName.escapePathSeparator()));
    await ignoreExceptions(file.delete);
  }

  /// Remove any cache entry that is not relevant given the remaining
  /// taskNames and keys.
  Future<void> removeNotMatching(
      Set<String> taskNames, Set<String> keys) async {
    var removedCount = 0;
    final oldTasks = Directory(_tasksDir).list();
    await for (final oldTask in oldTasks) {
      if (!taskNames.contains(path.basename(oldTask.path))) {
        logger.fine(() =>
            "Removing task '${oldTask.path}' directory as task is not part of the build");
        await oldTask.delete(recursive: true);
        removedCount++;
      }
    }
    final oldKeys = Directory(_hashesDir).list();
    await for (final oldKey in oldKeys) {
      if (!keys.contains(path.basename(oldKey.path))) {
        logger.fine(() =>
            "Removing key '${oldKey.path}' as key is not part of the build");
        await oldKey.delete(recursive: true);
        removedCount++;
      }
    }
    logger.fine(() =>
        'Removed $removedCount cache entries that are any longer relevant');
  }

  Future<void> _cacheFile(File file, {required String key}) async {
    final hf = _getCacheLocation(file, key: key);
    if (await file.exists()) {
      logger.finest(() => 'Caching file ${file.path} at ${hf.path}');
      await hf.parent.create(recursive: true);
      await hf.writeAsBytes((await hashFile(file)).bytes);
    } else {
      logger.finest(() =>
          'Removing file ${file.path} from ${hf.path} as it does not exist');
      await _removeEntity(file, key: key, cacheLocation: hf);
    }
  }

  Future<void> _cacheDir(Directory dir, Iterable<FileSystemEntity> children,
      {required String key}) async {
    final hf = _getCacheLocation(dir, key: key);

    logger.finest(() => 'Caching directory ${dir.path} at ${hf.path} with '
        'children $children');
    final contents = _DirectoryContents(children);
    await hf.parent.create(recursive: true);
    await hf.writeAsBytes(contents.encode());
  }

  Future<void> _removeEntity(FileSystemEntity entity,
      {required String key, File? cacheLocation}) async {
    final cl = cacheLocation ?? _getCacheLocation(entity, key: key);
    if (await cl.exists()) {
      logger.finest(() => 'Removing entry for ${entity.path}'
          ' with key $key from cache');
      await ignoreExceptions(cl.delete);
    }
  }

  /// Check if any member of a [FileCollection] has been modified since the
  /// last time someone checked with this method.
  ///
  /// The `key` argument is used to consider whether changes have happened
  /// since last time the check was made with the exact same key.
  ///
  /// Returns false if the [FileCollection] is empty.
  Future<bool> hasChanged(FileCollection fileCollection,
      {String key = ''}) async {
    return !await findChanges(fileCollection, key: key).isEmpty;
  }

  /// Find all changes if a [FileCollection] has been modified since the
  /// last time someone checked with this method.
  ///
  /// The `key` argument is used to consider whether changes have happened
  /// since last time the check was made with the exact same key.
  ///
  /// Returns an empty [Stream] if the [FileCollection] is empty.
  Stream<FileChange> findChanges(FileCollection fileCollection,
      {String key = ''}) async* {
    logger.finest(
        () => 'Checking if $fileCollection with key="$key" has changed');
    if (fileCollection.isEmpty) return;

    // if the cache is empty, avoid checking anything and report everything as added
    if (await isEmptyDir(_hashesDir)) {
      logger.fine('No hashes in the cache '
          '(${path.join(Directory.current.path, _hashesDir)}), '
          'reporting all as changed');
      yield* _reportAllAdded(fileCollection);
      return;
    }

    Set<String> visitedEntities = {};

    // visit all entities that currently exist
    await for (final entity in fileCollection.resolve()) {
      if (visitedEntities.add(entity.path)) {
        final changes = await entity.use((file) async* {
          final change = await _hasFileChanged(file, key: key);
          if (change != null) yield FileChange(file, change);
        }, (dir, children) async* {
          final dirChange = await _hasDirChanged(dir, children, key: key);
          if (dirChange != null && dirChange.kind == ChangeKind.modified) {
            // file additions/modifications are reported elsehwere,
            // but only here can we find deleted files.
            yield* _deletedFiles(
                dirChange.newContents!, dirChange.oldContents!);
          }
          if (dirChange != null) {
            yield FileChange(dir, dirChange.kind);
          }
        });
        yield* changes;
      }
    }

    // visit entities that do not exist but may have existed before
    for (final file in fileCollection.files.where(visitedEntities.add)) {
      final fileEntity = File(file);
      final change = await _hasFileChanged(fileEntity, key: key);
      if (change != null) {
        yield FileChange(fileEntity, change);
      }
    }
    for (final dir in fileCollection.directories
        .map((e) => e.path)
        .where(visitedEntities.add)) {
      // this dir doesn't exist, otherwise it would've been visited earlier
      final dirEntity = Directory(dir);
      final change = await _hasDirChanged(dirEntity, const [], key: key);
      if (change != null) {
        yield FileChange(dirEntity, change.kind);
      }
    }
  }

  Stream<FileChange> _reportAllAdded(FileCollection fileCollection) async* {
    await for (final entry in fileCollection.resolve()) {
      yield FileChange(entry.entity, ChangeKind.added);
    }
  }

  Future<ChangeKind?> _hasFileChanged(File file, {String key = ''}) async {
    final hf = _getCacheLocation(file, key: key);
    var hashExists = await hf.exists();
    if (!await file.exists()) {
      logger.fine(() => "File '${file.path}' does not exist "
          "${hashExists ? 'but was cached' : 'and was not known before'}");
      return hashExists ? ChangeKind.deleted : null;
    }
    ChangeKind? change;
    if (hashExists) {
      // allow for 1 second difference: file systems seem to not refresh
      // the timestamp with sub-second precision!
      if ((await file.lastModified())
          .add(const Duration(seconds: 1))
          .isAfter((await hf.lastModified()))) {
        logger.fine(
            () => "Detected possibly stale cache for file '${file.path}', "
                'checking file hash');
        final previousHash = await hf.readAsBytes();
        final hash = (await hashFile(file)).bytes;
        if (previousHash.equals(hash)) {
          logger.finest(
              () => "File '${file.path}' hash is still the same: '$hash'");
        } else {
          logger.fine(() => "File '${file.path}' hash changed - "
              "old hash='$previousHash', new hash='$hash'");
          change = ChangeKind.modified;
        }
      } else {
        logger.finest(() => "File '${file.path}' hash is fresh.");
      }
    } else {
      logger.fine(() => "Hash does not exist for file: '${file.path}'");
      change = ChangeKind.added;
    }
    return change;
  }

  Future<_DirectoryChange?> _hasDirChanged(
      Directory dir, Iterable<FileSystemEntity> children,
      {required String key}) async {
    final hf = _getCacheLocation(dir, key: key);
    _DirectoryChange? change;
    if (await hf.exists()) {
      final previousHash = await hf.readAsBytes();
      final currentContents = _DirectoryContents(children);
      final currentHash = currentContents.encode();
      if (previousHash.equals(currentHash)) {
        logger.finest(() => 'Directory hash is still the same: ${dir.path}');
      } else {
        logger.fine(() => 'Directory hash has changed: ${dir.path}');
        change = _DirectoryChange(ChangeKind.modified, currentContents,
            _DirectoryContents.decode(previousHash));
      }
    } else if (await dir.exists()) {
      logger.fine(() => 'Directory hash does not exist for '
          "existing directory: '${dir.path}'");
      change = _DirectoryChange(ChangeKind.added, null, null);
    } else {
      logger.finest(() => "Directory '${dir.path}' does not exist "
          'and was not known before');
    }
    return change;
  }

  File getExecutablesLocation(File file) =>
      File(path.join(_executablesDir, hash(file.path).toString()));

  File _getCacheLocation(FileSystemEntity entity, {required String key}) {
    final parentDir = entity is Directory ? entity.path : entity.parent.path;
    final fileName = _locationHash(entity);
    return File(path.join(
        _hashesDir, _encodeKey(key), parentDir.noPathNavigation(), fileName));
  }

  static String _locationHash(FileSystemEntity fe) =>
      hash(path.basename(fe.path)).toString();

  Stream<FileChange> _deletedFiles(
      _DirectoryContents newContents, _DirectoryContents oldContents) async* {
    final newEntities = newContents.children.map(entityPath).toSet();
    final deleted =
        oldContents.children.where((e) => !newEntities.contains(entityPath(e)));
    for (final entity in deleted) {
      yield FileChange(entity, ChangeKind.deleted);
    }
  }

  String _encodeKey(String key) {
    if (key.isEmpty) return key;
    return 'D__${key.escapePathSeparator()}__D';
  }
}

class _DirectoryChange {
  final ChangeKind kind;
  final _DirectoryContents? newContents;
  final _DirectoryContents? oldContents;

  _DirectoryChange(this.kind, this.newContents, this.oldContents);
}

/// Directory is stored in the cache as a list of its contents.
///
/// This is necessary because it must be able to detect file deletions,
/// so it must "remember" the previous files that were present directly under it.
class _DirectoryContents {
  final List<FileSystemEntity> children;

  _DirectoryContents(Iterable<FileSystemEntity> entities)
      : children = _sorted(entities);

  static List<FileSystemEntity> _sorted(Iterable<FileSystemEntity> entities) {
    final list = entities.toList(growable: false);
    list.sort((a, b) => a.path.compareTo(b.path));
    return list;
  }

  static _DirectoryContents decode(List<int> bytes) {
    final list = jsonDecode(utf8.decode(bytes)) as List;
    return _DirectoryContents(list.map((p) => fromEntityPath(p as String)));
  }

  List<int> encode() {
    final json = jsonEncode(children.map(entityPath).toList());
    return utf8.encode(json);
  }
}

final _pathSepPattern = RegExp(r'[\\/]');

extension _StringPaths on String {
  String escapePathSeparator() {
    return replaceAll(_pathSepPattern, r'$');
  }

  String noPathNavigation() {
    return replaceAll('..', r'$');
  }
}
